using DaggerfallConnect;
using DaggerfallConnect.Arena2;
using DaggerfallWorkshop.Game;
using DaggerfallWorkshop.Game.Entity;
using DaggerfallWorkshop.Game.Formulas;
using DaggerfallWorkshop.Game.UserInterface;
using DaggerfallWorkshop.Game.UserInterfaceWindows;
using System;
using System.Collections.Generic;
using UnityEngine;


namespace ViewableSkillProgress
{
    public class SkillsWindow : DaggerfallPopupWindow
    {
        PlayerEntity playerEntity = GameManager.Instance.PlayerEntity;

        Panel imagePanel = new Panel();
        Panel messagePanel = new Panel();
        Panel columnsPanel = new Panel();
        Panel column1Panel = new Panel();
        Panel column2Panel = new Panel();
        Panel columnSeparator = new Panel();


        public SkillsWindow(IUserInterfaceManager uiManager, IUserInterfaceWindow previous = null)
            : base(uiManager, previous)
        {
        }


        public void Show(List<DFCareer.Skills> skills, bool twoColumn = false)
        {
            if (!IsSetup)
                Setup();

            bool secondColumn = false;
            bool showHandToHandDamage = false;

            for (int i = 0; i < skills.Count; i++)
            {
                if (skills[i] == DFCareer.Skills.HandToHand)
                    showHandToHandDamage = true;

                if (!twoColumn)
                {
                    AddEntry(skills[i], column1Panel);
                }
                else
                {
                    AddEntry(skills[i], secondColumn ? column2Panel : column1Panel);
                    secondColumn = !secondColumn;
                }
            }

            if (showHandToHandDamage)
            {
                int minDamage = FormulaHelper.CalculateHandToHandMinDamage(playerEntity.Skills.GetLiveSkillValue(DFCareer.Skills.HandToHand));
                int maxDamage = FormulaHelper.CalculateHandToHandMaxDamage(playerEntity.Skills.GetLiveSkillValue(DFCareer.Skills.HandToHand));
                string text = DaggerfallUnity.TextProvider.GetSkillName(DFCareer.Skills.HandToHand) + " dmg: " + minDamage + "-" + maxDamage;

                AddEntry(text, column1Panel);
            }


            UpdatePanelSizes();

            if (column2Panel.Size.x == 0)
            {
                column1Panel.VerticalAlignment = VerticalAlignment.Middle;
                column1Panel.HorizontalAlignment = HorizontalAlignment.Center;
            }

            uiManager.PushWindow(this);

        }


        protected override void Setup()
        {
            if (IsSetup)
                return;

            base.Setup();

            allowFreeScaling = false;

            //---main message panel
            messagePanel.HorizontalAlignment = HorizontalAlignment.Center;
            messagePanel.VerticalAlignment = VerticalAlignment.Middle;

            DaggerfallUI.Instance.SetDaggerfallPopupStyle(DaggerfallUI.PopupStyle.Parchment, messagePanel);
            NativePanel.Components.Add(messagePanel);

            //---background image panel
            imagePanel.HorizontalAlignment = HorizontalAlignment.Center;
            imagePanel.VerticalAlignment = VerticalAlignment.Top;

            imagePanel.Position = new Vector2(0, 0);

            messagePanel.Components.Add(imagePanel);

            //---columns panel, contains column1 and column2
            columnsPanel.VerticalAlignment = VerticalAlignment.Middle;
            columnsPanel.HorizontalAlignment = HorizontalAlignment.Center;
            messagePanel.Components.Add(columnsPanel);

            //---column1 panel
            column1Panel.VerticalAlignment = VerticalAlignment.Top;
            column1Panel.HorizontalAlignment = HorizontalAlignment.Left;
            columnsPanel.Components.Add(column1Panel);

            //---columnSeparator, provides spacing between column1 and column2 if needed
            columnSeparator.HorizontalAlignment = HorizontalAlignment.Center;
            columnsPanel.Components.Add(columnSeparator);

            //---column2 panel
            column2Panel.VerticalAlignment = VerticalAlignment.Top;
            column2Panel.HorizontalAlignment = HorizontalAlignment.Right;
            columnsPanel.Components.Add(column2Panel);


            IsSetup = true;
        }


        public override void OnPush()
        {
            //basically a copy from DaggerfallMessageBox
            base.OnPush();
            parentPanel.OnMouseClick += ParentPanel_OnMouseClick;
            parentPanel.OnRightMouseClick += ParentPanel_OnMouseClick;
            parentPanel.OnMiddleMouseClick += ParentPanel_OnMouseClick;
        }

        public override void OnPop()
        {
            //basically a copy from DaggerfallMessageBox
            base.OnPop();
            parentPanel.OnMouseClick -= ParentPanel_OnMouseClick;
            parentPanel.OnRightMouseClick -= ParentPanel_OnMouseClick;
            parentPanel.OnMiddleMouseClick -= ParentPanel_OnMouseClick;
        }

        public override void Update()
        {
            base.Update();

            if (DaggerfallUI.Instance.HotkeySequenceProcessed == HotkeySequence.HotkeySequenceProcessStatus.NotFound)
            {
                if (Input.GetKeyUp(KeyCode.Return) || Input.GetKeyUp(KeyCode.KeypadEnter))
                {
                    CloseWindow();
                }
            }
        }


        public override void Draw()
        {
            base.Draw(); //here for debugging purposes
        }


        void UpdatePanelSizes()
        {
            Vector2 col1Size = CalculateSize(column1Panel);
            column1Panel.Size = col1Size;

            Vector2 col2Size = CalculateSize(column2Panel);
            column2Panel.Size = col2Size;

            const float columnSpacing = 10f;
            if (col2Size.x > 0) //only setting columnSeparator width if column2 has something in it
                columnSeparator.Size = new Vector2(columnSpacing, 0);

            float width = col1Size.x + col2Size.x + columnSeparator.Size.x;
            float height = Math.Max(col1Size.y, col2Size.y);

            columnsPanel.Size = new Vector2(width, height);

            // Resize the message panel to get a clean border of 22x22 pixel textures
            width += messagePanel.LeftMargin + messagePanel.RightMargin;
            height += messagePanel.TopMargin + messagePanel.BottomMargin;

            const int minimum = 44;

            if (width > minimum)
                width = (float)Math.Ceiling(width / 22) * 22;
            else
                width = minimum;

            if (height > minimum)
                height = (float)Math.Ceiling(height / 22) * 22;
            else
                height = minimum;

            messagePanel.Size = new Vector2(width, height);

        }


        Vector2 CalculateSize(Panel panel)
        {
            float maxX = 0;
            float maxY = 0;

            foreach (BaseScreenComponent component in panel.Components)
            {
                float x = component.Position.x + component.Size.x;
                float y = component.Position.y + component.Size.y;

                if (x > maxX)
                    maxX = x;

                if (y > maxY)
                    maxY = y;
            }

            return new Vector2(maxX, maxY);
        }


        void AddEntry(DFCareer.Skills skill, Panel column)
        {
            Panel entry = CreateEntry(skill);

            AddEntry(entry, column);
        }


        void AddEntry(string text, Panel column)
        {
            TextFile.Token token = new TextFile.Token();
            token.text = text;
            token.formatting = TextFile.Formatting.Text;

            Panel entry = new Panel();
            MultiFormatTextLabel label = new MultiFormatTextLabel();
            TextFile.Token[] tokens = new TextFile.Token[1];
            tokens[0] = token;
            label.SetText(tokens);
            entry.Components.Add(label);

            entry.Size = label.Size;

            AddEntry(entry, column);
        }


        void AddEntry(Panel entry, Panel column)
        {
            const float rowPadding = 2f;

            entry.Size = new Vector2(entry.InteriorWidth, entry.InteriorHeight);

            float y = 0;

            //determine y from last entry in the column
            foreach (BaseScreenComponent component in column.Components)
            {
                float nextY = component.Position.y + component.Size.y;
                if (nextY > y)
                    y = nextY;
            }

            if (y > 0)
                y += rowPadding;

            entry.Position = new Vector2(0f, y);

            column.Components.Add(entry);
        }


        Panel CreateEntry(DFCareer.Skills skill)
        {
            Panel panel = new Panel();
            
            TextFile.Token[] tokens = CreateSkillTokens(skill);

            MultiFormatTextLabel label = new MultiFormatTextLabel();
            label.HighlightColor = DaggerfallUI.DaggerfallUnityStatIncreasedTextColor;
            label.SetText(tokens);
            label.ToolTip = defaultToolTip;
            label.ToolTipText = "Add text for skill " + skill.ToString() + " here.";
            panel.Components.Add(label);

            Panel progressBar = CreateProgressBar(skill);
            panel.Components.Add(progressBar);

            float width = label.Size.x;
            float height = Math.Max(label.Size.y, progressBar.Size.y);
            panel.Size = new Vector2(width, height);

            return panel;
        }


        // Creates formatting tokens for skill popups
        TextFile.Token[] CreateSkillTokens(DFCareer.Skills skill)
        {
            bool highlight = playerEntity.GetSkillRecentlyIncreased(skill);

            List<TextFile.Token> tokens = new List<TextFile.Token>();
            TextFile.Formatting formatting = highlight ? TextFile.Formatting.TextHighlight : TextFile.Formatting.Text;

            TextFile.Token skillNameToken = new TextFile.Token();
            skillNameToken.formatting = formatting;
            skillNameToken.text = DaggerfallUnity.TextProvider.GetSkillName(skill);

            TextFile.Token skillValueToken = new TextFile.Token();
            skillValueToken.formatting = formatting;
            skillValueToken.text = string.Format("{0}", playerEntity.Skills.GetLiveSkillValue(skill));

            DFCareer.Stats primaryStat = DaggerfallSkills.GetPrimaryStat(skill);
            TextFile.Token skillPrimaryStatToken = new TextFile.Token();
            skillPrimaryStatToken.formatting = formatting;
            skillPrimaryStatToken.text = DaggerfallUnity.TextProvider.GetAbbreviatedStatName(primaryStat);

            TextFile.Token positioningToken = new TextFile.Token();
            positioningToken.formatting = TextFile.Formatting.PositionPrefix;

            TextFile.Token tabToken = new TextFile.Token();
            tabToken.formatting = TextFile.Formatting.PositionPrefix;

            tokens.Add(skillNameToken);
            int offset = DaggerfallWorkshop.DaggerfallUnity.Settings.SDFFontRendering ? 65 : 75;
            positioningToken.x = offset + ViewableSkillProgressMain.ProgressBarWidth;
            tokens.Add(positioningToken);
            tokens.Add(skillValueToken);

            if (ViewableSkillProgressMain.GovernAttributeText)
            {
                offset = DaggerfallWorkshop.DaggerfallUnity.Settings.SDFFontRendering ? 78 : 92;
                positioningToken.x = offset + ViewableSkillProgressMain.ProgressBarWidth;
                tokens.Add(positioningToken);
                tokens.Add(skillPrimaryStatToken);
            }

            return tokens.ToArray();
        }



        Panel CreateProgressBar(DFCareer.Skills skill)
        {
            float cT = CurrentTallyCount(skill);
            float aT = TallysNeededToAdvance(skill);
            if (cT > aT)
                cT = aT; // Continue work on this tomorrow, get the non-SDF stuff to work and then should be all good, if I don't plan to do the other feature as well at least.

            string progressText = GetProgressText(skill, cT, aT);

            Panel pan = new Panel();
            pan.Position = new Vector2(ViewableSkillProgressMain.ProgressBarPosX, 0);
            pan.Size = new Vector2(ViewableSkillProgressMain.ProgressBarWidth, 5);
            pan.ToolTip = defaultToolTip;
            pan.ToolTipText = progressText;

            /*
            int xPanModifier = DaggerfallUnity.Settings.SDFFontRendering ? -5 : 13;
            if (secondColumn)
                xPanModifier = DaggerfallUnity.Settings.SDFFontRendering ? 0 : 15;
            */

            if (ViewableSkillProgressMain.ProgressDisplayType == 0 || ViewableSkillProgressMain.ProgressDisplayType == 2) // Only show progress bar when "DisplayType" is 0 or 2
            {
                Panel bar = DaggerfallUI.AddPanel(new Rect(1, 0, ViewableSkillProgressMain.ProgressBarWidth, 5), pan);

                if (playerEntity.Skills.GetPermanentSkillValue(skill) >= 100)
                    bar.BackgroundColor = ViewableSkillProgressMain.MasteredBarColor; // black by default
                else if (playerEntity.AlreadyMasteredASkill() && playerEntity.Skills.GetPermanentSkillValue(skill) >= 95)
                    bar.BackgroundColor = ViewableSkillProgressMain.MaxedBarColor; // dark blue by default
                else if (cT >= aT)
                    bar.BackgroundColor = ViewableSkillProgressMain.ReadyToLevelBarColor; // dark green by default
                else
                {
                    bar = DaggerfallUI.AddPanel(new Rect(1, 0, (int)Mathf.Floor((cT / aT) * ViewableSkillProgressMain.ProgressBarWidth), 5), pan);
                    bar.BackgroundColor = ViewableSkillProgressMain.ProgressBarColor; // dark red by default
                }

                Outline panBorder = DaggerfallUI.AddOutline(new Rect(0, 0, ViewableSkillProgressMain.ProgressBarWidth, 5), ViewableSkillProgressMain.ProgressBarOutlineColor, pan); // yellow by default
            }

            if (ViewableSkillProgressMain.ProgressDisplayType == 0 || ViewableSkillProgressMain.ProgressDisplayType == 1) // Only show progress text when "DisplayType" is 0 or 1
            {
                TextLabel progText = DaggerfallUI.AddTextLabel(DaggerfallUI.DefaultFont, new Vector2(0, 1), string.Empty, pan);
                progText.HorizontalAlignment = HorizontalAlignment.Center;
                progText.TextScale = ViewableSkillProgressMain.ProgressTextScale;
                progText.TextColor = ViewableSkillProgressMain.ProgressTextColor;

                progText.Text = progressText;
            }

            return pan;
        }


        string GetProgressText(DFCareer.Skills skill, float cT, float aT)
        {
            string brL = GetBracketLeftType();
            string brR = GetBracketRightType(); // Just used for if user has the "TextBrackets" option enabled and set to something, otherwise returns empty string.

            float sT = cT; // Just here to make the "For Nerds" text display option cleaner to implement in this code.

            if (playerEntity.Skills.GetPermanentSkillValue(skill) >= 100 && ViewableSkillProgressMain.ProgressTextType != 2)
                return brL + "MASTERED" + brR;
            else if (playerEntity.AlreadyMasteredASkill() && playerEntity.Skills.GetPermanentSkillValue(skill) >= 95 && ViewableSkillProgressMain.ProgressTextType != 2)
                return brL + "MAXED" + brR;
            else if (cT >= aT && ViewableSkillProgressMain.ProgressTextType != 2)
                return brL + "READY" + brR;
            else
            {
                if (ViewableSkillProgressMain.ProgressTextType == 0)
                    return string.Format(brL + "{0} / {1}" + brR, cT, aT);
                else if (ViewableSkillProgressMain.ProgressTextType == 1)
                    return string.Format(brL + "{0}" + brR, aT - cT);
                else
                    return string.Format(brL + "{0} / {1}" + brR, sT, aT);
            }
        }


        int CurrentTallyCount(DFCareer.Skills skill)
        {
            int i = (int)skill;

            int reflexesMod = 0x10000 - (((int)playerEntity.Reflexes - 2) << 13);
            int calculatedSkillUses = (playerEntity.SkillUses[i] * reflexesMod) >> 16;

            return calculatedSkillUses;
        }


        int TallysNeededToAdvance(DFCareer.Skills skill)
        {
            int i = (int)skill;

            int skillAdvancementMultiplier = DaggerfallSkills.GetAdvancementMultiplier((DFCareer.Skills)i);
            float careerAdvancementMultiplier = playerEntity.Career.AdvancementMultiplier;
            int usesNeededForAdvancement = FormulaHelper.CalculateSkillUsesForAdvancement(playerEntity.Skills.GetPermanentSkillValue(i), skillAdvancementMultiplier, careerAdvancementMultiplier, playerEntity.Level);

            return usesNeededForAdvancement;
        }


        string GetBracketLeftType()
        {
            switch (ViewableSkillProgressMain.TextWithOrWithoutBrackets)
            {
                case 1: return "<";
                case 2: return "[";
                case 3: return "(";
                case 4: return "+";
                case 5: return "*";
                case 6: return "=";
                default: return "";
            }
        }


        string GetBracketRightType()
        {
            switch (ViewableSkillProgressMain.TextWithOrWithoutBrackets)
            {
                case 1: return ">";
                case 2: return "]";
                case 3: return ")";
                case 4: return "+";
                case 5: return "*";
                case 6: return "=";
                default: return "";
            }
        }


        void ParentPanel_OnMouseClick(BaseScreenComponent sender, Vector2 position)
        {
            if (uiManager.TopWindow == this)
            {
                CloseWindow();
            }
        }


    } //class

} //namespace
